@@ -4,6 +4,7 @@ Core utilities that are loaded into the root namespace or used internally.
 
 import re
 import sys
+import types
 import atexit
 import logging
 import importlib.resources
@@ -55,16 +56,84 @@ def error_message_hash(message):
     return hash(message)
 
 
+# We implement a custom enum class that's much simpler than Python's enum.Enum,
+# and simply maps to strings or ints. The enums are classes, so IDE's provide
+# autocompletion, and documenting with Sphinx is easy. That does mean we need a
+# metaclass though.
+
+
+class EnumType(type):
+    """Metaclass for enums and flags."""
+
+    def __new__(cls, name, bases, dct):
+        # Collect and check fields
+        member_map = {}
+        for key, val in dct.items():
+            if not key.startswith("_"):
+                val = key if val is None else val
+                if not isinstance(val, (int, str)):
+                    raise TypeError("Enum fields must be str or int.")
+                member_map[key] = val
+        # Some field values may have been updated
+        dct.update(member_map)
+        # Create class
+        klass = super().__new__(cls, name, bases, dct)
+        # Attach some fields
+        klass.__fields__ = tuple(member_map)
+        klass.__members__ = types.MappingProxyType(member_map)  # enums.Enum compat
+        # Create bound methods
+        for name in ["__dir__", "__iter__", "__getitem__", "__setattr__", "__repr__"]:
+            setattr(klass, name, types.MethodType(getattr(cls, name), klass))
+        return klass
+
+    def __dir__(cls):
+        # Support dir(enum). Note that this order matches the definition, but dir() makes it alphabetic.
+        return cls.__fields__
+
+    def __iter__(cls):
+        # Support list(enum), iterating over the enum, and doing ``x in enum``.
+        return iter([getattr(cls, key) for key in cls.__fields__])
+
+    def __getitem__(cls, key):
+        # Support enum[key]
+        return cls.__dict__[key]
+
+    def __repr__(cls):
+        if cls is BaseEnum:
+            return "<wgpu.utils.BaseEnum>"
+        pkg = cls.__module__.split(".")[0]
+        name = cls.__name__
+        options = []
+        for key in cls.__fields__:
+            val = cls[key]
+            options.append(f"'{key}' ({val})" if isinstance(val, int) else f"'{val}'")
+        return f"<{pkg}.{name} enum with options: {', '.join(options)}>"
+
+    def __setattr__(cls, name, value):
+        if name.startswith("_"):
+            super().__setattr__(name, value)
+        else:
+            raise RuntimeError("Cannot set values on an enum.")
+
+
+class BaseEnum(metaclass=EnumType):
+    """Base class for flags and enums.
+
+    Looks like Python's builtin Enum class, but is simpler; fields are simply ints or strings.
+    """
+
+    def __init__(self):
+        raise RuntimeError("Cannot instantiate an enum.")
+
+
 _flag_cache = {}  # str -> int
 
 
 def str_flag_to_int(flag, s):
     """Allow using strings for flags, i.e. 'READ' instead of wgpu.MapMode.READ.
-    No worries about repeated overhead, because the resuls are cached.
+    No worries about repeated overhead, because the results are cached.
     """
-    cache_key = (
-        f"{flag._name}.{s}"  # using private attribute, lets call this a friend func
-    )
+    cache_key = f"{flag.__name__}.{s}"  # use class name
     value = _flag_cache.get(cache_key, None)
 
     if value is None:
@@ -80,7 +149,7 @@ def str_flag_to_int(flag, s):
                 v = flag.__dict__[p.upper()]
                 value += v
             except KeyError:
-                raise ValueError(f"Invalid flag value for {flag}: '{p}'")
+                raise ValueError(f"Invalid flag value for {flag}: '{p}'") from None
         _flag_cache[cache_key] = value
 
     return value
@@ -89,7 +158,7 @@ def str_flag_to_int(flag, s):
 class ApiDiff:
     """Helper class to define differences in the API by annotating
     methods. This way, these difference are made explicit, plus they're
-    logged so we can automatically included these changes in the docs.
+    logged so we can automatically include these changes in the docs.
     """
 
     def __init__(self):
