@@ -4,7 +4,7 @@ The backend implementations of the wgpu API.
 
 import sys
 
-from ..classes import GPU as _base_GPU  # noqa
+from ..classes import GPU as _base_GPU  # noqa: N811
 
 
 def _register_backend(gpu):
@@ -14,7 +14,7 @@ def _register_backend(gpu):
 
     root_namespace = sys.modules["wgpu"].__dict__
     needed_attributes = (
-        "request_adapter",
+        "request_adapter_sync",
         "request_adapter_async",
         "wgsl_language_features",
     )
