@@ -12,11 +12,11 @@ from .. import _register_backend
 
 
 class GPU:
-    def request_adapter(self, **parameters):
+    def request_adapter_sync(self, **parameters):
         raise NotImplementedError("Cannot use sync API functions in JS.")
 
     async def request_adapter_async(self, **parameters):
-        gpu = window.navigator.gpu  # noqa
+        gpu = window.navigator.gpu  # noqa: F821
         return await gpu.request_adapter(**parameters)
 
     def get_preferred_canvas_format(self):
