@@ -1,20 +1,22 @@
-"""Utilities used in the wgpu-native backend.
-"""
+"""Utilities used in the wgpu-native backend."""
 
-import sys
 import ctypes
+import sys
+import time
+from queue import deque
 
-from ._ffi import ffi, lib
+import sniffio
+
+from ._ffi import ffi, lib, lib_path
 from ..._diagnostics import DiagnosticsBase
 from ...classes import (
     GPUError,
-    GPUOutOfMemoryError,
-    GPUValidationError,
-    GPUPipelineError,
     GPUInternalError,
+    GPUOutOfMemoryError,
+    GPUPipelineError,
+    GPUValidationError,
 )
 
-
 ERROR_TYPES = {
     "": GPUError,
     "OutOfMemory": GPUOutOfMemoryError,
@@ -29,8 +31,8 @@ if sys.platform.startswith("darwin"):
 
 
 def get_memoryview_and_address(data):
-    """Get a memoryview for the given data and its memory address.
-    The data object must support the buffer protocol.
+    """Get a memoryview and memory-address for the given data.
+    The data object must support the buffer protocol, and be contiguous.
     """
 
     # To get the address from a memoryview, there are multiple options.
@@ -56,9 +58,11 @@ def get_memoryview_and_address(data):
     m = memoryview(data)
 
     # Test that the data is contiguous.
+    # In most cases we'd want c_contiguous data, but the user may be
+    # playing fancy tricks so we check for general contiguous-ness only.
     # Note that pypy does not have the contiguous attribute, so we assume it is.
     if not getattr(m, "contiguous", True):
-        raise ValueError("The given texture data is not contiguous")
+        raise ValueError("The given data is not contiguous")
 
     # Get the address via ffi. In contrast to ctypes, this also
     # works for readonly data (e.g. bytes)
@@ -91,22 +95,16 @@ def get_wgpu_instance():
     return _the_instance
 
 
-def get_surface_id_from_canvas(canvas):
+def get_surface_id_from_info(present_info):
     """Get an id representing the surface to render to. The way to
     obtain this id differs per platform and GUI toolkit.
     """
 
-    # Use cached
-    surface_id = getattr(canvas, "_wgpu_surface_id", None)
-    if surface_id:
-        return surface_id
-
-    surface_info = canvas.get_surface_info()
-
     if sys.platform.startswith("win"):  # no-cover
+        GetModuleHandle = ctypes.windll.kernel32.GetModuleHandleW  # noqa: N806
         struct = ffi.new("WGPUSurfaceDescriptorFromWindowsHWND *")
-        struct.hinstance = ffi.NULL
-        struct.hwnd = ffi.cast("void *", int(surface_info["window"]))
+        struct.hinstance = ffi.cast("void *", GetModuleHandle(lib_path))
+        struct.hwnd = ffi.cast("void *", int(present_info["window"]))
         struct.chain.sType = lib.WGPUSType_SurfaceDescriptorFromWindowsHWND
 
     elif sys.platform.startswith("darwin"):  # no-cover
@@ -120,7 +118,7 @@ def get_surface_id_from_canvas(canvas):
         #     [ns_window.contentView setLayer:metal_layer];
         #     surface = wgpu_create_surface_from_metal_layer(metal_layer);
         # }
-        window = ctypes.c_void_p(surface_info["window"])
+        window = ctypes.c_void_p(present_info["window"])
 
         cw = ObjCInstance(window)
         try:
@@ -161,22 +159,22 @@ def get_surface_id_from_canvas(canvas):
         struct.chain.sType = lib.WGPUSType_SurfaceDescriptorFromMetalLayer
 
     elif sys.platform.startswith("linux"):  # no-cover
-        platform = surface_info.get("platform", "x11")
+        platform = present_info.get("platform", "x11")
         if platform == "x11":
             struct = ffi.new("WGPUSurfaceDescriptorFromXlibWindow *")
-            struct.display = ffi.cast("void *", surface_info["display"])
-            struct.window = int(surface_info["window"])
+            struct.display = ffi.cast("void *", present_info["display"])
+            struct.window = int(present_info["window"])
             struct.chain.sType = lib.WGPUSType_SurfaceDescriptorFromXlibWindow
         elif platform == "wayland":
             struct = ffi.new("WGPUSurfaceDescriptorFromWaylandSurface *")
-            struct.display = ffi.cast("void *", surface_info["display"])
-            struct.surface = ffi.cast("void *", surface_info["window"])
+            struct.display = ffi.cast("void *", present_info["display"])
+            struct.surface = ffi.cast("void *", present_info["window"])
             struct.chain.sType = lib.WGPUSType_SurfaceDescriptorFromWaylandSurface
         elif platform == "xcb":
             # todo: xcb untested
             struct = ffi.new("WGPUSurfaceDescriptorFromXcbWindow *")
-            struct.connection = ffi.cast("void *", surface_info["connection"])  # ??
-            struct.window = int(surface_info["window"])
+            struct.connection = ffi.cast("void *", present_info["connection"])  # ??
+            struct.window = int(present_info["window"])
             struct.chain.sType = lib.WGPUSType_SurfaceDescriptorFromXlibWindow
         else:
             raise RuntimeError("Unexpected Linux surface platform '{platform}'.")
@@ -188,25 +186,25 @@ def get_surface_id_from_canvas(canvas):
     surface_descriptor.label = ffi.NULL
     surface_descriptor.nextInChain = ffi.cast("WGPUChainedStruct *", struct)
 
-    surface_id = lib.wgpuInstanceCreateSurface(get_wgpu_instance(), surface_descriptor)
-
-    # Cache and return
-    canvas._wgpu_surface_id = surface_id
-    return surface_id
+    return lib.wgpuInstanceCreateSurface(get_wgpu_instance(), surface_descriptor)
 
 
-# The functions below are copied from codegen/utils.py
+# The functions below are copied from codegen/utils.py - let's keep these in sync!
 
 
-def to_snake_case(name):
+def to_snake_case(name, separator="_"):
     """Convert a name from camelCase to snake_case. Names that already are
     snake_case remain the same.
     """
     name2 = ""
     for c in name:
         c2 = c.lower()
-        if c2 != c and len(name2) > 0 and name2[-1] not in "_123":
-            name2 += "_"
+        if c2 != c and len(name2) > 0:
+            prev = name2[-1]
+            if c2 == "d" and prev in "123":
+                name2 = name2[:-1] + separator + prev
+            elif prev != separator:
+                name2 += separator
         name2 += c2
     return name2
 
@@ -218,7 +216,7 @@ def to_camel_case(name):
     is_capital = False
     name2 = ""
     for c in name:
-        if c == "_" and name2:
+        if c in "_-" and name2:
             is_capital = True
         elif is_capital:
             name2 += c.upper()
@@ -230,6 +228,79 @@ def to_camel_case(name):
     return name2
 
 
+async def async_sleep(delay):
+    """Async sleep that uses sniffio to be compatible with asyncio, trio, rendercanvas.utils.asyncadapter, and possibly more."""
+    libname = sniffio.current_async_library()
+    sleep = sys.modules[libname].sleep
+    await sleep(delay)
+
+
+class WgpuAwaitable:
+    """An object that can be waited for, either synchronously using sync_wait() or asynchronously using await.
+
+    The purpose of this class is to implememt the asynchronous methods in a
+    truely async manner, as well as to support a synchronous version of them.
+    """
+
+    def __init__(self, title, callback, finalizer, poll_function=None):
+        self.title = title  # for context in error messages
+        self.callback = callback  # only used to prevent it from being gc'd
+        self.finalizer = finalizer  # function to finish the result
+        self.poll_function = poll_function  # call this to poll wgpu
+        self.result = None
+
+    def set_result(self, result):
+        self.result = (result, None)
+
+    def set_error(self, error):
+        self.result = (None, error)
+
+    def _is_done(self):
+        self.poll_function()
+        return self.result is not None
+
+    def _finish(self):
+        try:
+            if not self.result:
+                raise RuntimeError(f"Waiting for {self.title} timed out.")
+            result, error = self.result
+            if error:
+                raise RuntimeError(error)
+            else:
+                return self.finalizer(result)
+        finally:
+            # Reset attrs to prevent potential memory leaks
+            self.callback = self.finalizer = self.poll_function = self.result = None
+
+    def sync_wait(self):
+        if self.result is not None:
+            pass
+        elif not self.poll_function:
+            raise RuntimeError("Expected callback to have already happened")
+        else:
+            while not self._is_done():
+                time.sleep(0)
+        return self._finish()
+
+    def __await__(self):
+        # There is no documentation on what __await__() is supposed to return, but we
+        # can certainly copy from a function that *does* know what to return
+        async def wait_for_callback():
+            # In all the async cases that I've tried, the result is either already set, or
+            # resolves after the first call to the poll function. To make sure that our
+            # sleep-logic actually works, we always do at least one sleep call.
+            await async_sleep(0)
+            if self.result is not None:
+                return
+            if not self.poll_function:
+                raise RuntimeError("Expected callback to have already happened")
+            while not self._is_done():
+                await async_sleep(0)
+
+        yield from wait_for_callback().__await__()
+        return self._finish()
+
+
 class ErrorHandler:
     """Object that logs errors, with the option to collect incoming
     errors elsewhere.
@@ -237,29 +308,38 @@ class ErrorHandler:
 
     def __init__(self, logger):
         self._logger = logger
-        self._proxy_stack = []
+        self._proxy_stack = deque()
+        self._proxy_messages = {}
         self._error_message_counts = {}
 
-    def capture(self, func):
-        """Send incoming error messages to the given func instead of logging them."""
-        self._proxy_stack.append(func)
+    def capture(self, name):
+        """Capture incoming error messages instead of logging them directly."""
+        self._proxy_stack.append(name)
 
-    def release(self, func):
-        """Release the given func."""
-        f = self._proxy_stack.pop(-1)
-        if f is not func:
+    def release(self, name):
+        """Release the given name, returning the last captured error."""
+        n = self._proxy_stack.pop()
+        if n is not name:
+            messages = [m for _, m in self._proxy_message.values()]
+            self._proxy_messages.clear()
             self._proxy_stack.clear()
-            self._logger.warning("ErrorHandler capture/release out of sync")
+            self._logger.error("ErrorHandler capture/release out of sync")
+            for message in messages:
+                self.log_error(message)
+        return self._proxy_messages.pop(name, None)
 
     def handle_error(self, error_type: str, message: str):
         """Handle an error message."""
         if self._proxy_stack:
-            self._proxy_stack[-1](error_type, message)
+            proxy_name = self._proxy_stack[-1]
+            if proxy_name in self._proxy_messages:
+                self.log_error(self._proxy_messages[proxy_name][1])
+            self._proxy_messages[proxy_name] = error_type, message
         else:
             self.log_error(message)
 
     def log_error(self, message):
-        """Hanle an error message by logging it, bypassing any capturing."""
+        """Handle an error message by logging it, bypassing any capturing."""
         # Get count for this message. Use a hash that does not use the
         # digits in the message, because of id's getting renewed on
         # each draw.
@@ -278,12 +358,11 @@ class ErrorHandler:
 
 class SafeLibCalls:
     """Object that copies all library functions, but wrapped in such
-    a way that errors occuring in that call are raised as exceptions.
+    a way that errors occurring in that call are raised as exceptions.
     """
 
     def __init__(self, lib, error_handler):
         self._error_handler = error_handler
-        self._error_message = None
         self._make_function_copies(lib)
 
     def _make_function_copies(self, lib):
@@ -293,27 +372,20 @@ class SafeLibCalls:
                 if callable(ob):
                     setattr(self, name, self._make_proxy_func(name, ob))
 
-    def _handle_error(self, error_type, message):
-        # If we already had an error, we log the earlier one now
-        if self._error_message:
-            self._error_handler.log_error(self._error_message[1])
-        # Store new error
-        self._error_message = (error_type, message)
-
     def _make_proxy_func(self, name, ob):
+        error_handler = self._error_handler
+
         def proxy_func(*args):
             # Make the call, with error capturing on
-            handle_error = self._handle_error
-            self._error_handler.capture(handle_error)
+            error_handler.capture(name)
             try:
                 result = ob(*args)
             finally:
-                self._error_handler.release(handle_error)
+                error_type_msg = error_handler.release(name)
 
             # Handle the error.
-            if self._error_message:
-                error_type, message = self._error_message
-                self._error_message = None
+            if error_type_msg is not None:
+                error_type, message = error_type_msg
                 cls = ERROR_TYPES.get(error_type, GPUError)
                 wgpu_error = cls(message)
                 # The line below will be the bottom line in the traceback,
