@@ -4,5 +4,5 @@ These can be accessed via ``wgpu.classes``,
 but are also available in the root wgpu namespace.
 """
 
-from ._classes import *  # noqa: F401, F403
+from ._classes import *  # noqa: F403
 from ._classes import __all__  # noqa: F401
