@@ -3,12 +3,10 @@ Code to provide a canvas to render to.
 """
 
 from . import _gui_utils  # noqa: F401
-from .base import WgpuCanvasInterface, WgpuCanvasBase, WgpuAutoGui  # noqa: F401
-from .offscreen import WgpuOffscreenCanvasBase  # noqa: F401
+from .base import WgpuCanvasInterface, WgpuCanvasBase, WgpuAutoGui
 
 __all__ = [
-    "WgpuCanvasInterface",
-    "WgpuCanvasBase",
     "WgpuAutoGui",
-    "WgpuOffscreenCanvasBase",
+    "WgpuCanvasBase",
+    "WgpuCanvasInterface",
 ]
