@@ -5,6 +5,21 @@ from collections import defaultdict
 from ._gui_utils import log_exception
 
 
+def create_canvas_context(canvas, render_methods):
+    """Create a GPUCanvasContext for the given canvas.
+
+    Helper function to keep the implementation of WgpuCanvasInterface
+    as small as possible.
+    """
+    backend_module = sys.modules["wgpu"].gpu.__module__
+    if backend_module == "wgpu._classes":
+        raise RuntimeError(
+            "A backend must be selected (e.g. with request_adapter()) before canvas.get_context() can be called."
+        )
+    CanvasContext = sys.modules[backend_module].GPUCanvasContext  # noqa: N806
+    return CanvasContext(canvas, render_methods)
+
+
 class WgpuCanvasInterface:
     """The minimal interface to be a valid canvas.
 
@@ -19,46 +34,62 @@ class WgpuCanvasInterface:
         super().__init__(*args, **kwargs)
         self._canvas_context = None
 
-    def get_surface_info(self):
-        """Get information about the native window / surface.
+    def get_present_methods(self):
+        """Get info on the present methods supported by this canvas.
 
-        This is used to obtain a surface id, so that wgpu can render to the
-        region of the screen occupied by the canvas. Should return None for
-        offscreen canvases. Otherwise, this should return a dict with a "window"
-        field. On Linux the dict should contain more fields, see the existing
-        implementations for reference.
+        Must return a small dict, used by the canvas-context to determine
+        how the rendered result will be presented to the canvas.
+        This method is only called once, when the context is created.
+
+        Each supported method is represented by a field in the dict. The value
+        is another dict with information specific to that present method.
+        A canvas backend must implement at least either "screen" or "bitmap".
+
+        With method "screen", the context will render directly to a surface
+        representing the region on the screen. The sub-dict should have a ``window``
+        field containing the window id. On Linux there should also be ``platform``
+        field to distinguish between "wayland" and "x11", and a ``display`` field
+        for the display id. This information is used by wgpu to obtain the required
+        surface id.
+
+        With method "bitmap", the context will present the result as an image
+        bitmap. On GPU-based contexts, the result will first be rendered to an
+        offscreen texture, and then downloaded to RAM. The sub-dict must have a
+        field 'formats': a list of supported image formats. Examples are "rgba-u8"
+        and "i-u8". A canvas must support at least "rgba-u8". Note that srgb mapping
+        is assumed to be handled by the canvas.
         """
-        return None
+        raise NotImplementedError()
 
     def get_physical_size(self):
         """Get the physical size of the canvas in integer pixels."""
         raise NotImplementedError()
 
-    def get_context(self, kind="webgpu"):
+    def get_context(self, kind="wgpu"):
         """Get the ``GPUCanvasContext`` object corresponding to this canvas.
 
         The context is used to obtain a texture to render to, and to
         present that texture to the canvas. This class provides a
         default implementation to get the appropriate context.
-
-        The ``kind`` argument is a remnant from the WebGPU spec and
-        must always be "webgpu".
         """
         # Note that this function is analog to HtmlCanvas.getContext(), except
         # here the only valid arg is 'webgpu', which is also made the default.
-        assert kind == "webgpu"
+        assert kind in ("wgpu", "webgpu", None)
         if self._canvas_context is None:
-            # Get the active wgpu backend module
-            backend_module = sys.modules["wgpu"].gpu.__module__
-            if backend_module == "wgpu._classes":
-                raise RuntimeError(
-                    "A backend must be selected (e.g. with request_adapter()) before canvas.get_context() can be called."
-                )
-            # Instantiate the context
-            CC = sys.modules[backend_module].GPUCanvasContext  # noqa: N806
-            self._canvas_context = CC(self)
+            render_methods = self.get_present_methods()
+            self._canvas_context = create_canvas_context(self, render_methods)
+
         return self._canvas_context
 
+    def present_image(self, image, **kwargs):
+        """Consume the final rendered image.
+
+        This is called when using the "bitmap" method, see ``get_present_methods()``.
+        Canvases that don't support offscreen rendering don't need to implement
+        this method.
+        """
+        raise NotImplementedError()
+
 
 class WgpuCanvasBase(WgpuCanvasInterface):
     """A convenient base canvas class.
@@ -77,11 +108,12 @@ class WgpuCanvasBase(WgpuCanvasInterface):
     also want to set ``vsync`` to False.
     """
 
-    def __init__(self, *args, max_fps=30, vsync=True, **kwargs):
+    def __init__(self, *args, max_fps=30, vsync=True, present_method=None, **kwargs):
         super().__init__(*args, **kwargs)
         self._last_draw_time = 0
         self._max_fps = float(max_fps)
         self._vsync = bool(vsync)
+        present_method  # noqa - We just catch the arg here in case a backend does implement support it
 
     def __del__(self):
         # On delete, we call the custom close method.
@@ -134,8 +166,15 @@ class WgpuCanvasBase(WgpuCanvasInterface):
         with log_exception("Draw error"):
             self.draw_frame()
         with log_exception("Present error"):
-            if self._canvas_context:
-                return self._canvas_context.present()
+            context = self._canvas_context
+            if context:
+                result = context.present()
+                method = result.pop("method")
+                if method == "bitmap":
+                    image = result.pop("data")
+                    self.present_image(image, **result)
+                else:
+                    pass  # method is "skip", "fail, ""screen"
 
     def _get_draw_wait_time(self):
         """Get time (in seconds) to wait until the next draw in order to honour max_fps."""
@@ -161,6 +200,10 @@ class WgpuCanvasBase(WgpuCanvasInterface):
         """Set the window size (in logical pixels)."""
         raise NotImplementedError()
 
+    def set_title(self, title):
+        """Set the window title."""
+        raise NotImplementedError()
+
     def close(self):
         """Close the window."""
         pass
@@ -193,7 +236,7 @@ class WgpuAutoGui:
         super().__init__(*args, **kwargs)
         self._last_event_time = 0
         self._pending_events = {}
-        self._event_handlers = defaultdict(set)
+        self._event_handlers = defaultdict(list)
 
     def _get_event_wait_time(self):
         """Calculate the time to wait for the next event dispatching.
@@ -208,7 +251,7 @@ class WgpuAutoGui:
     def _handle_event_rate_limited(
         self, event, call_later_func, match_keys, accum_keys
     ):
-        """Alternative `to handle_event()` for events that must be rate-limted.
+        """Alternative `to handle_event()` for events that must be rate-limited.
 
         If any of the ``match_keys`` keys of the new event differ from the currently
         pending event, the old event is dispatched now. The ``accum_keys`` keys of
@@ -270,18 +313,21 @@ class WgpuAutoGui:
         """
         # Collect callbacks
         event_type = event.get("event_type")
-        callbacks = self._event_handlers[event_type] | self._event_handlers["*"]
+        callbacks = self._event_handlers[event_type] + self._event_handlers["*"]
         # Dispatch
-        for callback in callbacks:
+        for _, callback in callbacks:
             with log_exception(f"Error during handling {event['event_type']} event"):
+                if event.get("stop_propagation", False):
+                    break
                 callback(event)
 
-    def add_event_handler(self, *args):
+    def add_event_handler(self, *args, order=0):
         """Register an event handler to receive events.
 
         Arguments:
             callback (callable): The event handler. Must accept a single event argument.
             *types (list of strings): A list of event types.
+            order (int): The order in which the handler is called. Lower numbers are called first. Default is 0.
 
         For the available events, see
         https://jupyter-rfb.readthedocs.io/en/stable/events.html.
@@ -327,7 +373,8 @@ class WgpuAutoGui:
 
         def decorator(_callback):
             for type in types:
-                self._event_handlers[type].add(_callback)
+                self._event_handlers[type].append((order, _callback))
+                self._event_handlers[type].sort(key=lambda x: x[0])
             return _callback
 
         if decorating:
@@ -342,4 +389,6 @@ class WgpuAutoGui:
             *types (list of strings): A list of event types.
         """
         for type in types:
-            self._event_handlers[type].remove(callback)
+            self._event_handlers[type] = [
+                (o, cb) for o, cb in self._event_handlers[type] if cb is not callback
+            ]
