@@ -9,6 +9,7 @@ or ``sudo apt install libglfw3-wayland`` when using Wayland.
 
 import sys
 import time
+import atexit
 import weakref
 import asyncio
 
@@ -34,9 +35,9 @@ if sys.platform.startswith("linux") and SYSTEM_IS_WAYLAND:
 
 
 # Some glfw functions are not always available
-set_window_content_scale_callback = lambda *args: None  # noqa: E731
-set_window_maximize_callback = lambda *args: None  # noqa: E731
-get_window_content_scale = lambda *args: (1, 1)  # noqa: E731
+set_window_content_scale_callback = lambda *args: None
+set_window_maximize_callback = lambda *args: None
+get_window_content_scale = lambda *args: (1, 1)
 
 if hasattr(glfw, "set_window_content_scale_callback"):
     set_window_content_scale_callback = glfw.set_window_content_scale_callback
@@ -103,13 +104,54 @@ KEY_MAP_MOD = {
 }
 
 
+def get_glfw_present_methods(window):
+    if sys.platform.startswith("win"):
+        return {
+            "screen": {
+                "platform": "windows",
+                "window": int(glfw.get_win32_window(window)),
+            }
+        }
+    elif sys.platform.startswith("darwin"):
+        return {
+            "screen": {
+                "platform": "cocoa",
+                "window": int(glfw.get_cocoa_window(window)),
+            }
+        }
+    elif sys.platform.startswith("linux"):
+        if is_wayland:
+            return {
+                "screen": {
+                    "platform": "wayland",
+                    "window": int(glfw.get_wayland_window(window)),
+                    "display": int(glfw.get_wayland_display()),
+                }
+            }
+        else:
+            return {
+                "screen": {
+                    "platform": "x11",
+                    "window": int(glfw.get_x11_window(window)),
+                    "display": int(glfw.get_x11_display()),
+                }
+            }
+    else:
+        raise RuntimeError(f"Cannot get GLFW surafce info on {sys.platform}.")
+
+
+def get_physical_size(window):
+    psize = glfw.get_framebuffer_size(window)
+    return int(psize[0]), int(psize[1])
+
+
 class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
     """A glfw window providing a wgpu canvas."""
 
     # See https://www.glfw.org/docs/latest/group__window.html
 
     def __init__(self, *, size=None, title=None, **kwargs):
-        ensure_app()
+        app.init_glfw()
         super().__init__(**kwargs)
 
         # Handle inputs
@@ -131,7 +173,7 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
         self._is_minimized = False
 
         # Register ourselves
-        all_glfw_canvases.add(self)
+        app.all_glfw_canvases.add(self)
 
         # Register callbacks. We may get notified too often, but that's
         # ok, they'll result in a single draw.
@@ -146,14 +188,15 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
         glfw.set_window_iconify_callback(self._window, weakbind(self._on_iconify))
 
         # User input
-        self._key_modifiers = []
-        self._pointer_buttons = []
+        self._key_modifiers = ()
+        self._pointer_buttons = ()
         self._pointer_pos = 0, 0
         self._double_click_state = {"clicks": 0}
         glfw.set_mouse_button_callback(self._window, weakbind(self._on_mouse_button))
         glfw.set_cursor_pos_callback(self._window, weakbind(self._on_cursor_pos))
         glfw.set_scroll_callback(self._window, weakbind(self._on_scroll))
         glfw.set_key_callback(self._window, weakbind(self._on_key))
+        glfw.set_char_callback(self._window, weakbind(self._on_char))
 
         # Initialize the size
         self._pixel_ratio = -1
@@ -185,7 +228,7 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             self._on_close()
 
     def _on_close(self, *args):
-        all_glfw_canvases.discard(self)
+        app.all_glfw_canvases.discard(self)
         if self._window is not None:
             glfw.destroy_window(self._window)  # not just glfw.hide_window
             self._window = None
@@ -211,8 +254,7 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
         # on some systems and in logical-pixels on other, we use the
         # framebuffer size and pixel ratio to derive the logical size.
         pixel_ratio = get_window_content_scale(self._window)[0]
-        psize = glfw.get_framebuffer_size(self._window)
-        psize = int(psize[0]), int(psize[1])
+        psize = get_physical_size(self._window)
 
         self._pixel_ratio = pixel_ratio
         self._physical_size = psize
@@ -264,32 +306,8 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
 
     # API
 
-    def get_surface_info(self):
-        if sys.platform.startswith("win"):
-            return {
-                "platform": "windows",
-                "window": int(glfw.get_win32_window(self._window)),
-            }
-        elif sys.platform.startswith("darwin"):
-            return {
-                "platform": "cocoa",
-                "window": int(glfw.get_cocoa_window(self._window)),
-            }
-        elif sys.platform.startswith("linux"):
-            if is_wayland:
-                return {
-                    "platform": "wayland",
-                    "window": int(glfw.get_wayland_window(self._window)),
-                    "display": int(glfw.get_wayland_display()),
-                }
-            else:
-                return {
-                    "platform": "x11",
-                    "window": int(glfw.get_x11_window(self._window)),
-                    "display": int(glfw.get_x11_display()),
-                }
-        else:
-            raise RuntimeError(f"Cannot get GLFW surafce info on {sys.platform}.")
+    def get_present_methods(self):
+        return get_glfw_present_methods(self._window)
 
     def get_pixel_ratio(self):
         return self._pixel_ratio
@@ -305,6 +323,9 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             raise ValueError("Window width and height must not be negative")
         self._set_logical_size((float(width), float(height)))
 
+    def set_title(self, title):
+        glfw.set_window_title(self._window, title)
+
     def _request_draw(self):
         if not self._request_draw_timer_running:
             self._request_draw_timer_running = True
@@ -338,12 +359,12 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             event_type = "pointer_down"
             buttons = set(self._pointer_buttons)
             buttons.add(button)
-            self._pointer_buttons = list(sorted(buttons))
+            self._pointer_buttons = tuple(sorted(buttons))
         elif action == glfw.RELEASE:
             event_type = "pointer_up"
             buttons = set(self._pointer_buttons)
             buttons.discard(button)
-            self._pointer_buttons = list(sorted(buttons))
+            self._pointer_buttons = tuple(sorted(buttons))
         else:
             return
 
@@ -352,9 +373,9 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             "x": self._pointer_pos[0],
             "y": self._pointer_pos[1],
             "button": button,
-            "buttons": list(self._pointer_buttons),
-            "modifiers": list(self._key_modifiers),
-            "ntouches": 0,  # glfw dows not have touch support
+            "buttons": tuple(self._pointer_buttons),
+            "modifiers": tuple(self._key_modifiers),
+            "ntouches": 0,  # glfw does not have touch support
             "touches": {},
         }
 
@@ -406,9 +427,9 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
                 "x": self._pointer_pos[0],
                 "y": self._pointer_pos[1],
                 "button": button,
-                "buttons": list(self._pointer_buttons),
-                "modifiers": list(self._key_modifiers),
-                "ntouches": 0,  # glfw dows not have touch support
+                "buttons": tuple(self._pointer_buttons),
+                "modifiers": tuple(self._key_modifiers),
+                "ntouches": 0,  # glfw does not have touch support
                 "touches": {},
             }
             self._handle_event_and_flush(ev)
@@ -425,9 +446,9 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             "x": self._pointer_pos[0],
             "y": self._pointer_pos[1],
             "button": 0,
-            "buttons": list(self._pointer_buttons),
-            "modifiers": list(self._key_modifiers),
-            "ntouches": 0,  # glfw dows not have touch support
+            "buttons": tuple(self._pointer_buttons),
+            "modifiers": tuple(self._key_modifiers),
+            "ntouches": 0,  # glfw does not have touch support
             "touches": {},
         }
 
@@ -443,8 +464,8 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             "dy": -100.0 * dy,
             "x": self._pointer_pos[0],
             "y": self._pointer_pos[1],
-            "buttons": list(self._pointer_buttons),
-            "modifiers": list(self._key_modifiers),
+            "buttons": tuple(self._pointer_buttons),
+            "modifiers": tuple(self._key_modifiers),
         }
         match_keys = {"modifiers"}
         accum_keys = {"dx", "dy"}
@@ -458,13 +479,13 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
             if modifier:
                 modifiers = set(self._key_modifiers)
                 modifiers.add(modifier)
-                self._key_modifiers = list(sorted(modifiers))
+                self._key_modifiers = tuple(sorted(modifiers))
         elif action == glfw.RELEASE:
             event_type = "key_up"
             if modifier:
                 modifiers = set(self._key_modifiers)
                 modifiers.discard(modifier)
-                self._key_modifiers = list(sorted(modifiers))
+                self._key_modifiers = tuple(sorted(modifiers))
         else:  # glfw.REPEAT
             return
 
@@ -486,18 +507,66 @@ class GlfwWgpuCanvas(WgpuAutoGui, WgpuCanvasBase):
         ev = {
             "event_type": event_type,
             "key": keyname,
-            "modifiers": list(self._key_modifiers),
+            "modifiers": tuple(self._key_modifiers),
         }
         self._handle_event_and_flush(ev)
 
+    def _on_char(self, window, char):
+        # Undocumented char event to make imgui work, see https://github.com/pygfx/wgpu-py/issues/530
+        ev = {
+            "event_type": "char",
+            "char_str": chr(char),
+            "modifiers": tuple(self._key_modifiers),
+        }
+        self._handle_event_and_flush(ev)
+
+    def present_image(self, image, **kwargs):
+        raise NotImplementedError()
+        # AFAIK glfw does not have a builtin way to blit an image. It also does
+        # not really need one, since it's the most reliable GUI backend to
+        # render to the screen.
+
 
 # Make available under a name that is the same for all gui backends
 WgpuCanvas = GlfwWgpuCanvas
 
 
-all_glfw_canvases = weakref.WeakSet()
-glfw._pygfx_mainloop = None
-glfw._pygfx_stop_if_no_more_canvases = False
+class AppState:
+    """Little container for state about the loop and glfw."""
+
+    def __init__(self):
+        self.all_glfw_canvases = weakref.WeakSet()
+        self._loop = None
+        self.stop_if_no_more_canvases = False
+        self._glfw_initialized = False
+
+    def init_glfw(self):
+        glfw.init()  # Safe to call multiple times
+        if not self._glfw_initialized:
+            self._glfw_initialized = True
+            atexit.register(glfw.terminate)
+
+    def get_loop(self):
+        if self._loop is None:
+            self._loop = self._get_loop()
+            self._loop.create_task(keep_glfw_alive())
+        return self._loop
+
+    def _get_loop(self):
+        try:
+            return asyncio.get_running_loop()
+        except Exception:
+            pass
+        try:
+            return asyncio.get_event_loop()
+        except RuntimeError:
+            pass
+        loop = asyncio.new_event_loop()
+        asyncio.set_event_loop(loop)
+        return loop
+
+
+app = AppState()
 
 
 def update_glfw_canvasses():
@@ -506,7 +575,7 @@ def update_glfw_canvasses():
     """
     # Note that _draw_frame_and_present already catches errors, it can
     # only raise errors if the logging system fails.
-    canvases = tuple(all_glfw_canvases)
+    canvases = tuple(app.all_glfw_canvases)
     for canvas in canvases:
         if canvas._need_draw and not canvas._is_minimized:
             canvas._need_draw = False
@@ -514,39 +583,50 @@ def update_glfw_canvasses():
     return len(canvases)
 
 
-async def mainloop():
-    loop = asyncio.get_event_loop()
+async def keep_glfw_alive():
+    """Co-routine that lives forever, keeping glfw going.
+
+    Although it stops the event-loop if there are no more canvases (and we're
+    running the loop), this task stays active and continues when the loop is
+    restarted.
+    """
+    # TODO: this is not particularly pretty. It'd be better to use normal asyncio to
+    # schedule draws and then also process events. But let's address when we do #355 / #391
     while True:
-        n = update_glfw_canvasses()
-        if glfw._pygfx_stop_if_no_more_canvases and not n:
-            break
         await asyncio.sleep(0.001)
         glfw.poll_events()
-    loop.stop()
-    glfw.terminate()
+        n = update_glfw_canvasses()
+        if app.stop_if_no_more_canvases and not n:
+            loop = asyncio.get_running_loop()
+            loop.stop()
 
 
-def ensure_app():
-    # It is safe to call init multiple times:
-    # "Additional calls to this function after successful initialization
-    # but before termination will return GLFW_TRUE immediately."
-    glfw.init()
-    if glfw._pygfx_mainloop is None:
-        loop = asyncio.get_event_loop()
-        glfw._pygfx_mainloop = mainloop()
-        loop.create_task(glfw._pygfx_mainloop)
+def poll_glfw_briefly(poll_time=0.1):
+    """Briefly poll glfw for a set amount of time.
+
+    Intended to work around the bug that destroyed windows sometimes hang
+    around if the mainloop exits: https://github.com/glfw/glfw/issues/1766
+
+    I found that 10ms is enough, but make it 100ms just in case. You should
+    only run this right after your mainloop stops.
+
+    """
+    end_time = time.perf_counter() + poll_time
+    while time.perf_counter() < end_time:
+        glfw.wait_events_timeout(end_time - time.perf_counter())
 
 
 def call_later(delay, callback, *args):
-    loop = asyncio.get_event_loop()
+    loop = app.get_loop()
     loop.call_later(delay, callback, *args)
 
 
 def run():
-    ensure_app()
-    loop = asyncio.get_event_loop()
-    if not loop.is_running():
-        glfw._pygfx_stop_if_no_more_canvases = True
-        loop.run_forever()
-    else:
-        pass  # Probably an interactive session
+    loop = app.get_loop()
+    if loop.is_running():
+        return  # Interactive mode!
+
+    app.stop_if_no_more_canvases = True
+    loop.run_forever()
+    app.stop_if_no_more_canvases = False
+    poll_glfw_briefly()
