@@ -1,14 +1,12 @@
 """
 Support for rendering in a Jupyter widget. Provides a widget subclass that
-can be used as cell output, or embedded in a ipywidgets gui.
+can be used as cell output, or embedded in an ipywidgets gui.
 """
 
-from collections import defaultdict
 import weakref
 import asyncio
 
-from .offscreen import WgpuOffscreenCanvasBase
-from .base import WgpuAutoGui
+from .base import WgpuAutoGui, WgpuCanvasBase
 
 import numpy as np
 from jupyter_rfb import RemoteFrameBuffer
@@ -18,18 +16,18 @@ from IPython.display import display
 pending_jupyter_canvases = []
 
 
-class JupyterWgpuCanvas(WgpuAutoGui, WgpuOffscreenCanvasBase, RemoteFrameBuffer):
+class JupyterWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, RemoteFrameBuffer):
     """An ipywidgets widget providing a wgpu canvas. Needs the jupyter_rfb library."""
 
     def __init__(self, *, size=None, title=None, **kwargs):
         super().__init__(**kwargs)
 
         # Internal variables
+        self._last_image = None
         self._pixel_ratio = 1
         self._logical_size = 0, 0
         self._is_closed = False
         self._request_draw_timer_running = False
-        self._event_handlers = defaultdict(set)
 
         # Register so this can be display'ed when run() is called
         pending_jupyter_canvases.append(weakref.ref(self))
@@ -56,9 +54,10 @@ class JupyterWgpuCanvas(WgpuAutoGui, WgpuOffscreenCanvasBase, RemoteFrameBuffer)
         self._request_draw_timer_running = False
         # The _draw_frame_and_present() does the drawing and then calls
         # present_context.present(), which calls our present() method.
-        # The resuls is either a numpy array or None, and this matches
+        # The result is either a numpy array or None, and this matches
         # with what this method is expected to return.
-        return self._draw_frame_and_present()
+        self._draw_frame_and_present()
+        return self._last_image
 
     # Implementation needed for WgpuCanvasBase
 
@@ -77,6 +76,9 @@ class JupyterWgpuCanvas(WgpuAutoGui, WgpuOffscreenCanvasBase, RemoteFrameBuffer)
         self.css_width = f"{width}px"
         self.css_height = f"{height}px"
 
+    def set_title(self, title):
+        pass  # not supported yet
+
     def close(self):
         RemoteFrameBuffer.close(self)
 
@@ -88,34 +90,14 @@ class JupyterWgpuCanvas(WgpuAutoGui, WgpuOffscreenCanvasBase, RemoteFrameBuffer)
             self._request_draw_timer_running = True
             call_later(self._get_draw_wait_time(), RemoteFrameBuffer.request_draw, self)
 
-    # Implementation needed for WgpuOffscreenCanvasBase
+    # Implementation needed for WgpuCanvasInterface
 
-    def present(self, texture):
-        # This gets called at the end of a draw pass via offscreen.GPUCanvasContext
-        device = texture._device
-        size = texture.size
-        bytes_per_pixel = 4
-        data = device.queue.read_texture(
-            {
-                "texture": texture,
-                "mip_level": 0,
-                "origin": (0, 0, 0),
-            },
-            {
-                "offset": 0,
-                "bytes_per_row": bytes_per_pixel * size[0],
-                "rows_per_image": size[1],
-            },
-            size,
-        )
-        return np.frombuffer(data, np.uint8).reshape(size[1], size[0], 4)
+    def get_present_methods(self):
+        return {"bitmap": {"formats": ["rgba-u8"]}}
 
-    def get_preferred_format(self):
-        # Use a format that maps well to PNG: rgba8norm. Use srgb for
-        # perseptive color mapping. This is the common colorspace for
-        # e.g. png and jpg images. Most tools (browsers included) will
-        # blit the png to screen as-is, and a screen wants colors in srgb.
-        return "rgba8unorm-srgb"
+    def present_image(self, image, **kwargs):
+        # Convert memoryview to ndarray (no copy)
+        self._last_image = np.frombuffer(image, np.uint8).reshape(image.shape)
 
 
 # Make available under a name that is the same for all gui backends
