@@ -8,32 +8,38 @@ import ctypes
 import importlib
 
 from .base import WgpuCanvasBase, WgpuAutoGui
-from ._gui_utils import get_alt_x11_display, get_alt_wayland_display, weakbind
-
-
-is_wayland = False  # We force Qt to use X11 in _gui_utils.py
+from ._gui_utils import (
+    logger,
+    SYSTEM_IS_WAYLAND,
+    get_alt_x11_display,
+    get_alt_wayland_display,
+    weakbind,
+    get_imported_qt_lib,
+)
 
 
 # Select GUI toolkit
-for libname in ("PySide6", "PyQt6", "PySide2", "PyQt5"):
-    if libname in sys.modules:
-        QtCore = importlib.import_module(".QtCore", libname)
-        QtWidgets = importlib.import_module(".QtWidgets", libname)
-        try:
-            WA_PaintOnScreen = QtCore.Qt.WidgetAttribute.WA_PaintOnScreen
-            WA_DeleteOnClose = QtCore.Qt.WidgetAttribute.WA_DeleteOnClose
-            PreciseTimer = QtCore.Qt.TimerType.PreciseTimer
-            KeyboardModifiers = QtCore.Qt.KeyboardModifier
-            FocusPolicy = QtCore.Qt.FocusPolicy
-            Keys = QtCore.Qt.Key
-        except AttributeError:
-            WA_PaintOnScreen = QtCore.Qt.WA_PaintOnScreen
-            WA_DeleteOnClose = QtCore.Qt.WA_DeleteOnClose
-            PreciseTimer = QtCore.Qt.PreciseTimer
-            KeyboardModifiers = QtCore.Qt
-            FocusPolicy = QtCore.Qt
-            Keys = QtCore.Qt
-        break
+libname, already_had_app_on_import = get_imported_qt_lib()
+if libname:
+    QtCore = importlib.import_module(".QtCore", libname)
+    QtGui = importlib.import_module(".QtGui", libname)
+    QtWidgets = importlib.import_module(".QtWidgets", libname)
+    try:
+        WA_PaintOnScreen = QtCore.Qt.WidgetAttribute.WA_PaintOnScreen
+        WA_DeleteOnClose = QtCore.Qt.WidgetAttribute.WA_DeleteOnClose
+        WA_InputMethodEnabled = QtCore.Qt.WidgetAttribute.WA_InputMethodEnabled
+        PreciseTimer = QtCore.Qt.TimerType.PreciseTimer
+        KeyboardModifiers = QtCore.Qt.KeyboardModifier
+        FocusPolicy = QtCore.Qt.FocusPolicy
+        Keys = QtCore.Qt.Key
+    except AttributeError:
+        WA_PaintOnScreen = QtCore.Qt.WA_PaintOnScreen
+        WA_DeleteOnClose = QtCore.Qt.WA_DeleteOnClose
+        WA_InputMethodEnabled = QtCore.Qt.WA_InputMethodEnabled
+        PreciseTimer = QtCore.Qt.PreciseTimer
+        KeyboardModifiers = QtCore.Qt
+        FocusPolicy = QtCore.Qt
+        Keys = QtCore.Qt
 else:
     raise ImportError(
         "Before importing wgpu.gui.qt, import one of PySide6/PySide2/PyQt6/PyQt5 to select a Qt toolkit."
@@ -129,17 +135,39 @@ def enable_hidpi():
 # needed for wgpu, so not our responsibility (some users may NOT want it set).
 enable_hidpi()
 
+_show_image_method_warning = (
+    "Qt falling back to offscreen rendering, which is less performant."
+)
+
 
 class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     """A QWidget representing a wgpu canvas that can be embedded in a Qt application."""
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, present_method=None, **kwargs):
         super().__init__(*args, **kwargs)
 
-        # Configure how Qt renders this widget
-        self.setAttribute(WA_PaintOnScreen, True)
-        self.setAttribute(WA_DeleteOnClose, True)
+        # Determine present method
+        self._surface_ids = self._get_surface_ids()
+        if not present_method:
+            self._present_to_screen = True
+            if SYSTEM_IS_WAYLAND:
+                # Trying to render to screen on Wayland segfaults. This might be because
+                # the "display" is not the real display id. We can tell Qt to use
+                # XWayland, so we can use the X11 path. This worked at some point,
+                # but later this resulted in a Rust panic. So, until this is sorted
+                # out, we fall back to rendering via an image.
+                self._present_to_screen = False
+        elif present_method == "screen":
+            self._present_to_screen = True
+        elif present_method == "bitmap":
+            self._present_to_screen = False
+        else:
+            raise ValueError(f"Invalid present_method {present_method}")
+
+        self.setAttribute(WA_PaintOnScreen, self._present_to_screen)
         self.setAutoFillBackground(False)
+        self.setAttribute(WA_DeleteOnClose, True)
+        self.setAttribute(WA_InputMethodEnabled, True)
         self.setMouseTracking(True)
         self.setFocusPolicy(FocusPolicy.StrongFocus)
 
@@ -151,21 +179,24 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
 
     def paintEngine(self):  # noqa: N802 - this is a Qt method
         # https://doc.qt.io/qt-5/qt.html#WidgetAttribute-enum  WA_PaintOnScreen
-        return None
+        if self._present_to_screen:
+            return None
+        else:
+            return super().paintEngine()
 
     def paintEvent(self, event):  # noqa: N802 - this is a Qt method
         self._draw_frame_and_present()
 
     # Methods that we add from wgpu (snake_case)
 
-    def get_surface_info(self):
+    def _get_surface_ids(self):
         if sys.platform.startswith("win") or sys.platform.startswith("darwin"):
             return {
                 "window": int(self.winId()),
             }
         elif sys.platform.startswith("linux"):
-            # The trick to use an al display pointer works for X11, but results in a segfault on Wayland ...
-            if is_wayland:
+            if False:
+                # We fall back to XWayland, see _gui_utils.py
                 return {
                     "platform": "wayland",
                     "window": int(self.winId()),
@@ -177,8 +208,21 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
                     "window": int(self.winId()),
                     "display": int(get_alt_x11_display()),
                 }
+
+    def get_present_methods(self):
+        global _show_image_method_warning
+        if self._surface_ids is None:
+            self._surface_ids = self._get_surface_ids()
+
+        methods = {}
+        if self._present_to_screen:
+            methods["screen"] = self._surface_ids
         else:
-            raise RuntimeError(f"Cannot get Qt surafce info on {sys.platform}.")
+            if _show_image_method_warning:
+                logger.warning(_show_image_method_warning)
+                _show_image_method_warning = None
+            methods["bitmap"] = {"formats": ["rgba-u8"]}
+        return methods
 
     def get_pixel_ratio(self):
         # Observations:
@@ -211,6 +255,9 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
             raise ValueError("Window width and height must not be negative")
         self.resize(width, height)  # See comment on pixel ratio
 
+    def set_title(self, title):
+        self.setWindowTitle(title)
+
     def _request_draw(self):
         if not self._request_draw_timer.isActive():
             self._request_draw_timer.start(int(self._get_draw_wait_time() * 1000))
@@ -224,11 +271,11 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     # User events to jupyter_rfb events
 
     def _key_event(self, event_type, event):
-        modifiers = [
+        modifiers = tuple(
             MODIFIERS_MAP[mod]
             for mod in MODIFIERS_MAP.keys()
             if mod & event.modifiers()
-        ]
+        )
 
         ev = {
             "event_type": event_type,
@@ -237,26 +284,40 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
         }
         self._handle_event_and_flush(ev)
 
+    def _char_input_event(self, char_str):
+        ev = {
+            "event_type": "char",
+            "char_str": char_str,
+            "modifiers": None,
+        }
+        self._handle_event_and_flush(ev)
+
     def keyPressEvent(self, event):  # noqa: N802
         self._key_event("key_down", event)
+        self._char_input_event(event.text())
 
     def keyReleaseEvent(self, event):  # noqa: N802
         self._key_event("key_up", event)
 
+    def inputMethodEvent(self, event):  # noqa: N802
+        commit_string = event.commitString()
+        if commit_string:
+            self._char_input_event(commit_string)
+
     def _mouse_event(self, event_type, event, touches=True):
         button = BUTTON_MAP.get(event.button(), 0)
-        buttons = [
+        buttons = tuple(
             BUTTON_MAP[button]
             for button in BUTTON_MAP.keys()
             if button & event.buttons()
-        ]
+        )
 
         # For Qt on macOS Control and Meta are switched
-        modifiers = [
+        modifiers = tuple(
             MODIFIERS_MAP[mod]
             for mod in MODIFIERS_MAP.keys()
             if mod & event.modifiers()
-        ]
+        )
 
         ev = {
             "event_type": event_type,
@@ -296,16 +357,16 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
 
     def wheelEvent(self, event):  # noqa: N802
         # For Qt on macOS Control and Meta are switched
-        modifiers = [
+        modifiers = tuple(
             MODIFIERS_MAP[mod]
             for mod in MODIFIERS_MAP.keys()
             if mod & event.modifiers()
-        ]
-        buttons = [
+        )
+        buttons = tuple(
             BUTTON_MAP[button]
             for button in BUTTON_MAP.keys()
             if button & event.buttons()
-        ]
+        )
 
         ev = {
             "event_type": "wheel",
@@ -332,6 +393,38 @@ class QWgpuWidget(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     def closeEvent(self, event):  # noqa: N802
         self._handle_event_and_flush({"event_type": "close"})
 
+    def present_image(self, image_data, **kwargs):
+        size = image_data.shape[1], image_data.shape[0]  # width, height
+
+        painter = QtGui.QPainter(self)
+
+        # We want to simply blit the image (copy pixels one-to-one on framebuffer).
+        # Maybe Qt does this when the sizes match exactly (like they do here).
+        # Converting to a QPixmap and painting that only makes it slower.
+
+        # Just in case, set render hints that may hurt performance.
+        painter.setRenderHints(
+            painter.RenderHint.Antialiasing | painter.RenderHint.SmoothPixmapTransform,
+            False,
+        )
+
+        image = QtGui.QImage(
+            image_data,
+            size[0],
+            size[1],
+            size[0] * 4,
+            QtGui.QImage.Format.Format_RGBA8888,
+        )
+
+        rect1 = QtCore.QRect(0, 0, size[0], size[1])
+        rect2 = self.rect()
+        painter.drawImage(rect2, image, rect1)
+
+        # Uncomment for testing purposes
+        # painter.setPen(QtGui.QColor("#0000ff"))
+        # painter.setFont(QtGui.QFont("Arial", 30))
+        # painter.drawText(100, 100, "This is an image")
+
 
 class QWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     """A toplevel Qt widget providing a wgpu canvas."""
@@ -341,11 +434,12 @@ class QWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     # size can be set to subpixel (logical) values, without being able to
     # detect this. See https://github.com/pygfx/wgpu-py/pull/68
 
-    def __init__(self, *, size=None, title=None, max_fps=30, **kwargs):
+    def __init__(
+        self, *, size=None, title=None, max_fps=30, present_method=None, **kwargs
+    ):
         # When using Qt, there needs to be an
         # application before any widget is created
         get_app()
-
         super().__init__(**kwargs)
 
         self.setAttribute(WA_DeleteOnClose, True)
@@ -353,7 +447,9 @@ class QWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
         self.setWindowTitle(title or "qt wgpu canvas")
         self.setMouseTracking(True)
 
-        self._subwidget = QWgpuWidget(self, max_fps=max_fps)
+        self._subwidget = QWgpuWidget(
+            self, max_fps=max_fps, present_method=present_method
+        )
         self._subwidget.add_event_handler(weakbind(self.handle_event), "*")
 
         # Note: At some point we called `self._subwidget.winId()` here. For some
@@ -384,8 +480,8 @@ class QWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     def draw_frame(self, f):
         self._subwidget.draw_frame = f
 
-    def get_surface_info(self):
-        return self._subwidget.get_surface_info()
+    def get_present_methods(self):
+        return self._subwidget.get_present_methods()
 
     def get_pixel_ratio(self):
         return self._subwidget.get_pixel_ratio()
@@ -401,6 +497,9 @@ class QWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
             raise ValueError("Window width and height must not be negative")
         self.resize(width, height)  # See comment on pixel ratio
 
+    def set_title(self, title):
+        self.setWindowTitle(title)
+
     def _request_draw(self):
         return self._subwidget._request_draw()
 
@@ -419,6 +518,9 @@ class QWgpuCanvas(WgpuAutoGui, WgpuCanvasBase, QtWidgets.QWidget):
     def request_draw(self, *args, **kwargs):
         return self._subwidget.request_draw(*args, **kwargs)
 
+    def present_image(self, image, **kwargs):
+        return self._subwidget.present_image(image, **kwargs)
+
 
 # Make available under a name that is the same for all gui backends
 WgpuWidget = QWgpuWidget
@@ -431,7 +533,12 @@ def get_app():
 
 
 def run():
+    if already_had_app_on_import:
+        return  # Likely in an interactive session or larger application that will start the Qt app.
     app = get_app()
+
+    # todo: we could detect if asyncio is running (interactive session) and wheter we can use QtAsyncio.
+    # But let's wait how things look with new scheduler etc.
     app.exec() if hasattr(app, "exec") else app.exec_()
 
 
